;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;;    this file is part of CavityStuffer - a molecular polymer design tool
;;;                         -------------
;;; copyright (c) 1993 by Institute for Molecular Manufacturing, Palo Alto, CA
;;; copyright (c) 1994 and beyond: M. Krummenacker (KR) kr@n-a-n-o.com
;;;
;;; -This program is free software; you can redistribute it and/or modify
;;;   it under the terms of the GNU General Public License as published by
;;;   the Free Software Foundation.
;;; -This program is distributed in the hope that it will be useful, but WITHOUT
;;;   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
;;;   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU GPL for more details.
;;; -You should have received a copy of the GNU General Public License along with
;;;   this program (the file "COPYING-2.0"); if not, write to the Free Software
;;;   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; for customization for various LISP implementations, the following
;;; conditionalized read symbols are used below:
;;;  :CMU GCL ECL CLISP MCL

(when (and (find-package "COMMON-LISP")
           (find-symbol "DEFPACKAGE" (find-package "COMMON-LISP"))
           )
  (defpackage "CAVSTUF"
     (:nicknames "CVST")
     (:use "COMMON-LISP" "COMMON-LISP-USER")
     (:export "LOAD-CAVSTUF" "COMPILE-CAVSTUF")
     )
  )
;;; in MCL, see below for (shadow 'read-line)

;;; for all the old-fashioned prehistoric LISPs that do not have the new
;;; packages as required by CLtL2
(unless (find-package "COMMON-LISP")
  (make-package "COMMON-LISP"
                :nicknames '("CL")
                :use '("LISP")
                )
  )
(unless (find-package "COMMON-LISP-USER")
  (make-package "COMMON-LISP-USER"
                :nicknames '("CL-USER")
                :use '("LISP")
                )
  )

;;; probably the (in-package ...) below takes care of this already, as it
;;; ought to implicitly (make-package ...) in the old pre-CLtL2 LISPs
#+(or GCL ECL)
(unless (find-package "CAVSTUF")
  (make-package "CAVSTUF"
                :nicknames '("CVST")
                :use '( #-GCL "COMMON-LISP"      #+GCL "LISP"
                        #-GCL "COMMON-LISP-USER" #+GCL "USER")
                )
  )

;;; leave behind our mark.
(pushnew :CAVSTUF *features*)

;;;kr98-04-18: these are convenient for debugging. however, they are kind
;;; of verbose, and the slow-down is noticeable.
;;; they are active during compile time. so uncomment the lines if needed,
;;; before building cavstuf.
;;;(pushnew :VERBOSE-SUCCESS-MESSAGES *features*)
;;;(pushnew :VERBOSE-CLASH-MESSAGES *features*)
;;;(pushnew :VERBOSE-GROWTH *features*)

(in-package "CAVSTUF")

;;;kr97-02-03 these bozos have not put this variable in the right package.
#+(or GCL ECL)
(unless (find-symbol "*LOAD-PATHNAME*")
  (import 'system:*load-pathname*)
  )
  
#| kr97-02-03 let's try something better
#-(or GCL ECL)  ;; these crappy old LISPs still do not seem to have this :-(
(setf *default-pathname-defaults*
      (make-pathname :directory (pathname-directory *LOAD-PATHNAME*)
		     ))
|#

;;;kr97-02-03 let's try something better
(defvar *cavstuf-root-dir* (pathname-directory (truename *load-pathname*)))

#|;;; should we use something like this ?   kr 95-9-11
(def-logical-pathname "modules-root" 
  (directory-namestring *loading-file-source-file*))
;;; or this, from zebu ??
(let ((*default-pathname-defaults*
       (make-pathname :directory (pathname-directory *LOAD-PATHNAME*)
		      :type "lisp"))
|#

;;; used in the (read-line ...) replacement below
#+MCL
(defun string-insert-extend (new-element
                             vector
                             &optional (position (length vector)))
  "new-element
   vector              ; has to be :adjustable, but no :fillpointer!!
   &optional position  ; default if not supplied is the length of the vector
Takes the new-element and inserts it into the vector at the specified
position (counting from zero, as usual). Adjusts the length of the vector
beforehand if the position lies outside the previous length."
  (if (>= position (length vector))
    (setf vector (adjust-array vector
                               (+ 1 position)
                             ;;  :initial-element nil
                               )))
  (setf (elt vector position) new-element))

;; insert setting *warn-if-redefine-kernel* to nil for the following in MCL:
;; (kr 95-7-28) (but see below for something much better!)
;;; use (shadow ...) to avoid the conflict. kr 95-10-10
#+MCL
(shadow 'read-line)

;;; replace (read-line ...) with a look-alike that allows slurping up files
;;;  with Mac CR as well as UN!X LF
;;;;not implemented yet: possibly a problem: treat eof-error-p properly!!!!!
;;;kr96-6-7: ok, should work now.
;;;
#+MCL
(defun read-line (&optional (input-stream *standard-input*)
                            eof-error-p
                            eof-value
                            recursive-p)
  "input-stream
   eof-error-p
   eof-value
   recursive-p
An emulation which tolerates new-line characters from UN!X, Mac, and even PC clones.
The resulting string does not contain a new-line character.
Has an enhancement: returns as the third value the type of the new-line read."
  (let ((result-string (make-array 0
                                   :element-type 'character
                                   :adjustable t
                                 ;;  :initial-element nil
                                   ))
        (eof-flag nil)
        (line-type-flag nil))
    (do ((current-character)
         (line-end-flag nil))
        (line-end-flag                    ; end-test: return the string
         (values result-string
                 eof-flag
                 line-type-flag))         ; return some interesting info
      (setf current-character
            (read-char input-stream nil 'eof recursive-p))
      (case current-character
        ('eof
         (progn
           (setf eof-flag t)              ; abnormal line termination as a result
           (setf line-end-flag t)         ; let the end-test terminate this line
           (setf line-type-flag 'eof)
           (when (= 0 (length result-string))
             ;; eof-error-p handling happens only when 'eof occurs
             ;; immediately at the beginning of a line.
             (if eof-error-p
               (error "eof occurred on empty line on ~A~%." input-stream)
               ;; otherwise, return eof-value
               (setf result-string eof-value)
               )
             )
           ))
  ;;      (#\Newline             ; generic Common LISP (actually this is erronous !)
  ;;       (setf end-flag t))    ; let the end-test terminate this line
        (#\Linefeed                       ; UN!X
         (progn
           (setf line-end-flag t)         ; let the end-test terminate this line
           (setf line-type-flag 'un!x)))
        (#\Return                         ; Mac, and possibly even a screwy PC clone
         (let ((one-character-buffer
                (read-char input-stream nil 'eof recursive-p)))
           (if (equal one-character-buffer #\Linefeed)
             ;; now we are truly faced with one of these PC clones
             (progn
               (setf line-end-flag t)     ; let the end-test terminate this line
               (setf line-type-flag 'ibm))
             ;; it's just a Mac: restore this fetched look-ahead again
             (progn
               (unless (eq 'eof one-character-buffer)
                 (unread-char one-character-buffer input-stream))
               (setf line-end-flag t)     ; let the end-test terminate this line
               (setf line-type-flag 'mac)))))
        (otherwise (string-insert-extend current-character result-string)))
      )))
#| just a few tests
(setf file (open "MacPartition:cubane.gnu.pdb"))
(setf file (open "MacPartition:cubane.mac.pdb"))
(setf file (open "MacPartition:cubane.ibm.pdb"))
(setf file (open "MacPartition:cubane.trunc.pdb"))
(read-line file)
(read-char file nil 'eof)
|#

;;; set some directory paths

(defparameter PDB-moiety-directory
  (append *cavstuf-root-dir*
          ;; sorted in order of increasingly brain-dead lack of features
          #-(or :CMU17 GCL ECL CLISP)
          '(:relative :up  "PDB_files" "PDBmoieties0.1")
          #+CLISP
          '(:relative ".." "PDB_files" "PDBmoieties0.1")
          #+(or :CMU17 GCL ECL)
          '(          ".." "PDB_files" "PDBmoieties0.1")
          )
  "This file-directory is used by the PDB-parsing routines 
for reading moieties while constructing the moiety library.")

(defparameter PDB-CavStuf-directory 
  (append *cavstuf-root-dir*
          ;; sorted in order of increasingly brain-dead lack of features
          #-(or :CMU17 GCL ECL CLISP)
          '(:relative :up  "PDB_files" "PDBCavStuf")
          #+CLISP
          '(:relative ".." "PDB_files" "PDBCavStuf")
          #+(or :CMU17 GCL ECL)
          '(          ".." "PDB_files" "PDBCavStuf")
          )
  "This file-directory is used by the PDB-parsing routines 
for reading moieties while constructing the moiety library.")

(defparameter Lib-directory
  (append *cavstuf-root-dir*
          ;; sorted in order of increasingly brain-dead lack of features
          #-(or :CMU17 GCL ECL CLISP)
          '(:relative :up  "CavStuf-Libraries")
          #+CLISP
          '(:relative ".." "CavStuf-Libraries")
          #+(or :CMU17 GCL ECL)
          '(          ".." "CavStuf-Libraries")
          )
  "This file-directory is used by the Moiety-type-Library
managing routines.")

(defvar version-string "cavstuf-0.3"
  "This string defines the version number of CavStuf 
and is referenced during writing to files, for example.")

(defparameter *write-space-saving-pdb-files* t
  "This flag to suppress writing white-space to fill precisely 80 columns 
in PDB files. Default is turned on. Turn off with nil.")

(defvar *cavstuf-files* (list
                         ;; load these first, as they can contain macros and defstructs
                         "utils"
                         "structure-defs"
                         ;; the order of the following does not matter as much
                         "matrix-vector"
                         "norm-map"
                         "file-library"
                         "pdb-in-out"
                         "create-moiety-type"
                         "create-link-type"
                         "clash"
                         "tree-anchorage"
                         "tree-growth"
                         "tree-dump"
                         "growth-driver"
                         "biased-random"
                         "atm-properties"
                         "tree-shorthand"
                         ) ; list
  )

;;; load in all the relevant files
(defun load-cavstuf-source ()
  (dolist (file *cavstuf-files*)
    (load (make-pathname :directory *cavstuf-root-dir*
                         :name file
                         :type "lisp"))
    )
  (format t "~&;;; ~A loaded" version-string)
  )

;;; load in all the relevant files.
;;;kr98-04-18: now loads binary files. obviously should be made a lot smarter.
;;;
(defun load-cavstuf ()
  #+(or GCL ECL)
  (load (make-pathname :directory *cavstuf-root-dir*
                       :name (format nil "~A-temp" #+GCL "gcl" #+ECL "ecl")
                       :type "o") )

  #-(or GCL ECL)
  (dolist (file *cavstuf-files*)
    (load (make-pathname :directory *cavstuf-root-dir*
                         :name file
                         :type #+(or GCL ECL) "o"
                               #+(and :CMU17 :X86) "x86f"
                               ;; is there any reasonable default ??
                               #-(or :CMU17 GCL ECL) "lisp"
                               ))
    )
  (format t "~&;;; ~A loaded" version-string)
  )

(defun compile-cavstuf ()
  "first loads the source, then compiles, and then loads the binaries."
  ;;kr98-04-18: we probably should load all the source first, 'cos of macros
  (load-cavstuf-source)
  ;; these two lisps benefit from having all the code in
  ;; just one file, because it enables the compiler to
  ;; better optimize function calls.
  #+(or GCL ECL)
  (let ((result-string "")
        (temp-file (format nil "~A-temp.lisp" #+GCL "gcl" #+ECL "ecl"))
        )
    (system (format nil "cat ~A >~A"
                    (dolist (file *cavstuf-files* result-string)
                      (setf result-string
                            (concatenate 'string result-string file ".lisp "))
                      )
                    temp-file
                    )
            )
    (compile-file temp-file)
    )

  #-(or GCL ECL)
  (dolist (file *cavstuf-files*)
    ;;;(compile-file (concatenate 'string file ".lisp"))
    ;;kr98-01-26: let's do this like in (load-cavstuf )
    (compile-file (make-pathname :directory *cavstuf-root-dir*
                                 :name file
                                 :type "lisp"))
    )

  (fresh-line)
  (format t "~&;;; ~A compiled" version-string)
  ;; let's load it too
  (load-cavstuf)
  )

