;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;;    this file is part of CavityStuffer - a molecular polymer design tool
;;;                         -------------
;;; copyright (c) 1993 by Institute for Molecular Manufacturing, Palo Alto, CA
;;; copyright (c) 1994 and beyond: M. Krummenacker (KR) kr@n-a-n-o.com
;;;
;;; -This program is free software; you can redistribute it and/or modify
;;;   it under the terms of the GNU General Public License as published by
;;;   the Free Software Foundation.
;;; -This program is distributed in the hope that it will be useful, but WITHOUT
;;;   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
;;;   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU GPL for more details.
;;; -You should have received a copy of the GNU General Public License along with
;;;   this program (the file "COPYING-2.0"); if not, write to the Free Software
;;;   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(in-package "CAVSTUF")

(defstruct atm-characteristics
  MM2-vdW-epsilon  ; in tJ
  MM2-vdW-r        ; in pm
  )

(defvar atm-properties nil
  "Denotes a list of atm-characteristics structures.")

;; for the MM2 case, the parameters are:
(setf atm-properties
      (vector (; atm-type 0  no proper atom
               make-atm-characteristics
               :MM2-vdW-epsilon   '?
               :MM2-vdW-r         0)
              (; atm-type 1  H
               make-atm-characteristics
               :MM2-vdW-epsilon   0.382
               :MM2-vdW-r         150)
              (; atm-type 2  C
               make-atm-characteristics
               :MM2-vdW-epsilon   0.357
               :MM2-vdW-r         190)
              (; atm-type 3  N
               make-atm-characteristics
               :MM2-vdW-epsilon   0.447
               :MM2-vdW-r         182)
              (; atm-type 4  O
               make-atm-characteristics
               :MM2-vdW-epsilon   0.536  ; the carbonyl variety
               :MM2-vdW-r         174)
              (; atm-type 5  F
               make-atm-characteristics
               :MM2-vdW-epsilon   0.634
               :MM2-vdW-r         165)
              (; atm-type 6  Si
               make-atm-characteristics
               :MM2-vdW-epsilon   1.137
               :MM2-vdW-r         225)
              (; atm-type 7  P
               make-atm-characteristics
               :MM2-vdW-epsilon   1.365
               :MM2-vdW-r         218)
              (; atm-type 8  S
               make-atm-characteristics
               :MM2-vdW-epsilon   1.641
               :MM2-vdW-r         211)
              (; atm-type 9  Cl
               make-atm-characteristics
               :MM2-vdW-epsilon   1.950
               :MM2-vdW-r         203)
              (; atm-type 10 Br
               make-atm-characteristics
               :MM2-vdW-epsilon   2.599
               :MM2-vdW-r         218)
              ))

(defun atm-MM2-vdW-radii ()
  "no parameters
returns a list."
  (map 'list #'(lambda (atm-description)
                 (atm-characteristics-MM2-vdW-r atm-description))
       atm-properties))

(defun atm-MM2-0.1nN-radii ()
  "no parameters
returns a list."
  (map 'list #'(lambda (atm-description)
                 (loaded-MM2-radius
                  0.1
                  (atm-characteristics-MM2-vdW-epsilon atm-description)
                  (atm-characteristics-MM2-vdW-r atm-description)))
       atm-properties))

(defun atm-MM2-1.0nN-radii ()
  "no parameters
returns a list."
  (map 'list #'(lambda (atm-description)
                 (loaded-MM2-radius
                  1.0
                  (atm-characteristics-MM2-vdW-epsilon atm-description)
                  (atm-characteristics-MM2-vdW-r atm-description)))
       atm-properties))

(defun atm-MM2-1.5nN-radii ()
  "no parameters
returns a list."
  (map 'list #'(lambda (atm-description)
                 (loaded-MM2-radius
                  1.5
                  (atm-characteristics-MM2-vdW-epsilon atm-description)
                  (atm-characteristics-MM2-vdW-r atm-description)))
       atm-properties))

(defun loaded-MM2-radius (nN-force MM2-vdW-epsilon MM2-vdW-r)
  "nN-force         ; in nN
   MM2-vdW-epsilon  ; in tJ
   MM2-vdW-r        ; in pm
This function implements Eq. (3.20) in 'Nanosystems'"
  (if (= MM2-vdW-r 0.0)
    0.0
    (let ((force (/ nN-force 1E9))
          (vdW-epsilon (/ MM2-vdW-epsilon 1E21))
          (vdW-r (/ MM2-vdW-r 1E12)))
      (* 1E12
         (/ vdW-r 13.6)
         (log (/ (* 2.5E6 vdW-epsilon)
                 (* vdW-r force)))))))
