 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;;    this file is part of CavityStuffer - a molecular polymer design tool
;;;                         -------------
;;; copyright (c) 1993 by Institute for Molecular Manufacturing, Palo Alto, CA
;;; copyright (c) 1994 and beyond: M. Krummenacker (KR) kr@n-a-n-o.com
;;;
;;; -This program is free software; you can redistribute it and/or modify
;;;   it under the terms of the GNU General Public License as published by
;;;   the Free Software Foundation.
;;; -This program is distributed in the hope that it will be useful, but WITHOUT
;;;   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
;;;   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU GPL for more details.
;;; -You should have received a copy of the GNU General Public License along with
;;;   this program (the file "COPYING-2.0"); if not, write to the Free Software
;;;   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(in-package "CAVSTUF")

(defun parse-pdb-file (file-name file-format)
  "file-name ;file as a string, including extension
   file-format ;the format translation-table
Reads-in a file that specifies the atom coordinates and connectivity
of a moiety-type. Returns raw-format, which is a list of two vectors,
the first containing parsed-atms, the second parsed-bond-starlets.
This an alternative, more detailed version of the pdb-parser."
  (let ((atm-vector  (make-array 0 :adjustable t
                                 ;;  :fill-pointer t
                                   ))
        (bond-starlet-vector (make-array 0 :adjustable t
                                           :fill-pointer t))
        (one-line-string ""))
    (with-open-file (in-filestream 
                     (make-pathname :directory PDB-moiety-directory
                                    :name file-name)
                     :direction :input)
      (do ((line-counter 1 (1+ line-counter))
           (end-of-file-flag nil))
          (end-of-file-flag (error "EOF occurred within line ~D"
                                   line-counter))
        (multiple-value-setq
         (one-line-string
          end-of-file-flag)
         (read-line in-filestream))
;                    nil
;                    (error "EOF occurred in line ~D,\
;         before valid terminating line was encountered" line-counter)))
        (multiple-value-bind (atm bonds control)
                             (parse-pdb-line one-line-string
                                             file-format
                                             line-counter)
          (ecase control
                 ;;kr96-10-30: let's try to do this better.
                 ;; will allow array-like access to atms.
                 ;;(atm (vector-push-extend atm atm-vector))
                 (atm (sequence-insert-extend atm atm-vector :index (1- (first atm)) ))

                 (bonds (vector-push-extend bonds bond-starlet-vector))
                 (atms&bonds
                  ;;kr96-10-30: let's try to do this better.
                  ;; will allow array-like access to atms.
                  ;;(vector-push-extend atm atm-vector)
                  (sequence-insert-extend atm atm-vector :index (1- (first atm)) )

                  (vector-push-extend bonds bond-starlet-vector))
                 (not-recognized nil)
                 (end-reached (return (generate-raw-format
                                          atm-vector
                                          bond-starlet-vector)))))))))

(defun parse-pdb-line (one-line-string translation-table line-number)
  "one-line-string   ;on line of a file
   translation-table ;the format translation-table
   line-number       ;number of line in file, for error information
Works in close conjunction with read&parse-all-lines-of-a-file.
Returns three values, the last of which acts as a control channel.
A large part of the parsing instructions are actually contained
in the translation-table, within of which the returned values are generated
whenever the first word of a line is recognized"
  (multiple-value-bind (operator found-flag)
           (gethash (string-right-trim Whitespace-filter
                                       (subseq one-line-string 0 
                                               (if (<
                                                    (length one-line-string)
                                                    6)
                                                 (length one-line-string)
                                                 6)))
                    translation-table)
     (if (eql found-flag nil)
         (values nil nil 'not-recognized)
         (if (equal (first operator) "x")
           (funcall (second operator)                ; executable directly
                  one-line-string translation-table line-number)
           (funcall (eval (second operator))         ; needs to be compiled first
                  one-line-string translation-table line-number)))))

#|
(progn (setf pdb-parse
             (read&parse-all-lines-of-a-file "norbornyl.pdb" pdb-Chem3D+))
       (setf parsed-atm-vector (first pdb-parse))
       (setf bond-vector (second pdb-parse)))
(read&parse-all-lines-of-a-file "trash.pdb" pdb-Chem3D+)
(read&parse-all-lines-of-a-file "norbornyl.pdb" pdb-Chem3D+)
(read&parse-all-lines-of-a-file "trispikepoly.pdb" pdb-Chem3D+)
(parse-line "COMPND  Norbornyl                                         "                         
            pdb-Chem3D+ 1)
(parse-line
"HETATM    1  O                 -1.0556  0.0677  0.1841                "
            pdb-Chem3D+ 3)
(parse-line
"HETATM  228  C       A 228      28.219  27.256  23.069                "
            pdb-Chem3D+ 3)
(parse-line "CONECT    7    1    4   18   19     "                         
            pdb-Chem3D+ 2)
chokes with an error if integers are malformed:
(parse-line "CONECT screwed-up    1    4   18   19     "                         
            pdb-Chem3D+ 3)
(parse-line "CONECT   12    3                                             "                         
            pdb-Chem3D+ 4)
(parse-line
"CONECT    7   11"
            pdb-Chem3D+ 3)
(parse-line "END                         "                         
            pdb-Chem3D+ 5)
(parse-line "END"                         
            pdb-Chem3D+ 5)
|#
;;; Construct an initial minimal translation-table pdb-bnl
(progn (setf pdb-bnl (make-hash-table :test #'equal))
        (setf (gethash 'initial-state pdb-bnl) '(tbd))
        (setf (gethash "CONECT" pdb-bnl)
              '("s" #'(lambda (one-line-string dummy1 dummy2)
                        (declare (ignore dummy1 dummy2))
                        (let ((center-atm
                               (parse-integer one-line-string
                                              :start 6 :end 11))
                              (line-length (length one-line-string)))
                          (do ((end-of-line-flag nil)
                               (temp-atm nil)
                               (index 11)
                               (bond-ends-list '()))
                              (end-of-line-flag
                               (if (eql bond-ends-list '())
                                 (values nil
                                         nil
                                         'not-recognized)
                                 (values nil
                                         ;; put together a bond-starlet:
                                         (list center-atm bond-ends-list)
                                         'bonds)))
                            (multiple-value-setq
                             (temp-atm
                              index)                       ;;index gets advanced
                             (parse-integer one-line-string
                                            :start index 
                                            :end (+ index 5)
                                            :junk-allowed T))
                            (if (eql temp-atm nil)         ;;Test if integer
                              (setf end-of-line-flag T)    ;;  was (malformed)
                              (setf bond-ends-list         ;;  or non-existant
                                    (nconc bond-ends-list (list temp-atm))))
                            (if (>= index line-length)
                              (setf end-of-line-flag T)
                              nil))))))
        (setf (gethash "HETATM" pdb-bnl)
              '("s" #'(lambda (one-line-string translation-table line-number)
                        (let* ((atm-number
                                (parse-integer one-line-string
                                               :start 6 :end 11))
                               (atm-label
                                (string-trim Whitespace-filter
                                             ;; this can pick up (junk) numbers
                                             ;; try leaving this out for a while
                                             ;;(subseq one-line-string 12 16)
                                             (subseq one-line-string 12 14)
                                             ))
                               (xyz (vector
                                     (* 100 (read-from-string
                                             one-line-string T nil
                                             :start 30 :end 38))
                                     (* 100 (read-from-string
                                             one-line-string T nil
                                             :start 38 :end 46))
                                     (* 100 (read-from-string
                                             one-line-string T nil
                                             :start 46 :end 54)))))
                          (multiple-value-bind (atm-kind found-flag)
                                               (gethash atm-label
                                                        translation-table)
                            (if (eql found-flag nil)
;;nil  ;kr96-09-29: let's reinstate this stuff for real
                                (warn "unknown label ~S for atom number ~A ; in line ~D~%"
                                      atm-label atm-number line-number)
                                (values (list atm-number
                                              (make-atm :xyz  xyz
                                                        :type (gethash atm-kind
                                                                       translation-table)))
                                        nil
                                        'atm)))))))
        (setf (gethash "ATOM" pdb-bnl)
              '("s" #'(lambda (one-line-string translation-table line-number)
                        (let* ((atm-number
                                (parse-integer one-line-string
                                               :start 6 :end 11))
                               (atm-label
                                (string-trim Whitespace-filter
                                             ;; this can pick up (junk) numbers
                                             ;; try leaving this out for a while
                                             ;;(subseq one-line-string 12 16)
                                             (subseq one-line-string 12 14)
                                             ))
                               (xyz (vector
                                     (* 100 (read-from-string
                                             one-line-string T nil
                                             :start 30 :end 38))
                                     (* 100 (read-from-string
                                             one-line-string T nil
                                             :start 38 :end 46))
                                     (* 100 (read-from-string
                                             one-line-string T nil
                                             :start 46 :end 54)))))
                          (multiple-value-bind (atm-kind found-flag)
                                               (gethash atm-label
                                                        translation-table)
                            (if (eql found-flag nil)
;;nil  ;kr96-09-29: let's reinstate this stuff for real
                                (warn "unknown label ~S for atom number ~A ; in line ~D~%"
                                      atm-label atm-number line-number)
                                (values (list atm-number
                                              (make-atm :xyz  xyz
                                                        :type (gethash atm-kind
                                                                       translation-table)))
                                        nil
                                        'atm)))))))
        (setf (gethash "END" pdb-bnl)
              '("s" #'(lambda (dummy1 dummy2 dummy3)
                        (declare (ignore dummy1 dummy2 dummy3))
                        (values nil
                                nil
                                'end-reached))))
        ;; special weirdness for msi
        (setf (gethash "H_" pdb-bnl) 'hydrogen)
        (setf (gethash "C_" pdb-bnl) 'carbon)
        (setf (gethash "N_" pdb-bnl) 'nitrogen)
        (setf (gethash "O_" pdb-bnl) 'oxygen)
        (setf (gethash "S_" pdb-bnl) 'sulfur)
        ;; more regular stuff
        (setf (gethash "H" pdb-bnl) 'hydrogen)
        (setf (gethash "C" pdb-bnl) 'carbon)
        (setf (gethash "N" pdb-bnl) 'nitrogen)
        (setf (gethash "O" pdb-bnl) 'oxygen)
        (setf (gethash "F" pdb-bnl) 'fluorine)
        (setf (gethash "Si" pdb-bnl) 'silicon)
        (setf (gethash "P" pdb-bnl) 'phosphorous)
        (setf (gethash "S" pdb-bnl) 'sulfur)
        (setf (gethash "Cl" pdb-bnl) 'chlorine)
        (setf (gethash "Br" pdb-bnl) 'bromine)
        (setf (gethash 'hydrogen pdb-bnl) 1)
        (setf (gethash 'carbon pdb-bnl) 2)
        (setf (gethash 'nitrogen pdb-bnl) 3)
        (setf (gethash 'oxygen pdb-bnl) 4)
        (setf (gethash 'fluorine pdb-bnl) 5)
        (setf (gethash 'silicon pdb-bnl) 6)
        (setf (gethash 'phosphorous pdb-bnl) 7)
        (setf (gethash 'sulfur pdb-bnl) 8)
        (setf (gethash 'chlorine pdb-bnl) 9)
        (setf (gethash 'bromine pdb-bnl) 10)
        (values t ;;(dump-hash-table pdb-bnl)
                pdb-bnl))

;;;kr96-8-31
;;; based on FIXED SLOTS.
;;; generalized the (string-insert-extend ) and (vector-insert-extend ) from older cavstuf.
;;; see CLtL2 p. 454 for a discussion of fill-pointers and p. 456 for (adjust-array ) .
;;;
(defun sequence-insert-extend (new-element
                               sequence
                               &key index
                               )
  "new-element
   sequence          ; has to be :adjustable. will be destructively modified.
   &key index        ; if not supplied, the default is the length of the sequence
Takes the new-element and inserts it into the sequence at the specified
position (counting from zero, as usual). Adjusts the length of the sequence
beforehand if the position of :index lies outside the previous length.
The position is treated as a fixed slot, which is overwritten."
  (let* ((len (length sequence))
         (position (if index
                       index
                       len) )
         )
    (when (>= position len)
          (setf sequence (adjust-array sequence (+ 1 position) ) )
          )
    (setf (elt sequence position) new-element)
    position
    ) ; let*
  )

#|
(let ((str (make-array 4 :element-type 'character :adjustable t :fill-pointer 0) ))
  (sequence-insert-extend #\a str)
  str
  )
(let ((lis (make-array 4 :adjustable t) ))
  (sequence-insert-extend 'A lis :index 10)
  lis
  )
|#
