;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;;    this file is part of CavityStuffer - a molecular polymer design tool
;;;                         -------------
;;; copyright (c) 1993 by Institute for Molecular Manufacturing, Palo Alto, CA
;;; copyright (c) 1994 and beyond: M. Krummenacker (KR) kr@n-a-n-o.com
;;;
;;; -This program is free software; you can redistribute it and/or modify
;;;   it under the terms of the GNU General Public License as published by
;;;   the Free Software Foundation.
;;; -This program is distributed in the hope that it will be useful, but WITHOUT
;;;   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
;;;   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU GPL for more details.
;;; -You should have received a copy of the GNU General Public License along with
;;;   this program (the file "COPYING-2.0"); if not, write to the Free Software
;;;   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(in-package "CAVSTUF")

;;; The tree-structure containing the actual molecular stuffing structure
;;; is the value of a symbol that names the tree and this symbol
;;; has a few properties attached:
;;;  - volume        (in nm^3)
;;;  - clash-array ???(not yet; still global)
;;;  - open-sites    (see "growth-driver"; used during growing the tree)
;;;  - ...(whatever) (maybe the shorthand tree structure ?)


(defvar Current-tree nil
  "Denotes the currently active stuffing-tree")

(defun get-current-tree-symbol ()
  "no parameters
Obtains the symbol which harbors the currently active stuffing-tree."
  (values Current-tree))

(defun get-current-tree ()
  "no parameters
Obtains the currently active stuffing-tree"
  (values (symbol-value Current-tree)
          Current-tree))

(defun set-current-tree (tree)
  "tree
Sets the currently active stuffing-tree to the tree-structure supplied by tree"
  (setf (symbol-value Current-tree) tree))

(defun init-current-tree-volume ()
  "no parameters
Initializes the volume of the currently active stuffing-tree"
  (setf (get (get-current-tree-symbol) 'volume) 0.0))

(defun get-current-tree-volume ()
  "no parameters
Obtains the volume of the currently active stuffing-tree"
  (get (get-current-tree-symbol) 'volume))

(defun set-current-tree-volume (volume)
  "volume
Sets the volume of the currently active stuffing-tree to the supplied value"
  (setf (get (get-current-tree-symbol) 'volume) volume))

(defun get-current-tree-clash-array ()
  "no parameters
Obtains the volume of the currently active stuffing-tree"
  (get (get-current-tree-symbol) 'clash-array))

(defun set-current-tree-clash-array (clash-array)
  "volume
Sets the volume of the currently active stuffing-tree to the supplied value"
  (setf (get (get-current-tree-symbol) 'clash-array) clash-array))

