;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;;    this file is part of CavityStuffer - a molecular polymer design tool
;;;                         -------------
;;; copyright (c) 1993 by Institute for Molecular Manufacturing, Palo Alto, CA
;;; copyright (c) 1994 and beyond: M. Krummenacker (KR) kr@n-a-n-o.com
;;;
;;; -This program is free software; you can redistribute it and/or modify
;;;   it under the terms of the GNU General Public License as published by
;;;   the Free Software Foundation.
;;; -This program is distributed in the hope that it will be useful, but WITHOUT
;;;   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
;;;   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU GPL for more details.
;;; -You should have received a copy of the GNU General Public License along with
;;;   this program (the file "COPYING-2.0"); if not, write to the Free Software
;;;   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(in-package "CAVSTUF")


;;; a general language-addition :
(defmacro dosequence ((var sequenceform &optional resultform)
                       &body body)
"Generalizes dolist to any kind of sequence."
  (let ((position #-:DEBUG-GENSYM (make-symbol "" ;;cl-user:*zilch-string*
                                               )
                  #+:DEBUG-GENSYM (gensym "DOSEQUENCE") ))
    `(let ((seq-len (length ,sequenceform)))
       (do ((,position 0 (1+ ,position)))
           ((>= ,position seq-len)
            (let ((,var nil))
              ,var                 ;use it trivially.
              ,resultform))
         (let ((,var (elt ,sequenceform ,position)))
           ,@body)))))

#| found this in ~/lisp/gcl/gcl-2.2.1/cmpnew/cmputil.lsp
   i do not understand the reason why temp needs to be a gensym
   in any case, above, it should be possible to also declare fixnum,
   and probably the position could be declared dynamic-extent.
(defmacro dotimes** ((v n &optional (val nil)) . body)
  (let ((temp (gensym)))
   `(do* ((,temp ,n) (,v 0 (1+ ,v)))
	 ((>= ,v ,temp) ,val)
	 (declare (fixnum ,temp ,v))
	 ,@body)))
|#

(defmacro doseq-with-index ((var
                             index
                             sequenceform
                             &optional resultform)
                            &body body)
"Generalizes dolist to any kind of sequence.
And it caries along with it the counting index."
  `(do ((,index 0 (1+ ,index)))
       ((>= ,index (length ,sequenceform))
        (let ((,var nil))
          ,var                 ;use it trivially.
          ,resultform))
     (let ((,var (elt ,sequenceform ,index)))
       ,@body)))



(defun nice-time-stamp ()
  "no args
Returns the full current time as a nice looking stamp, in a string of length 19,
in the format \"1996-07-04 17:07:50\"."
  (multiple-value-bind (second minute hour date month year
                       day-of-week daylight-savings-p time-zone)
                       (get-decoded-time)
    (declare (ignore day-of-week daylight-savings-p time-zone))
    (format nil "~4,'0D-~2,'0D-~2,'0D ~2,'0D:~2,'0D:~2,'0D"
            year month date hour minute second)
    )
  )

(defun dump-hash-table (hash-table)
  (let ((dump-list (make-array 0 :adjustable t
                                 :element-type 'cons
                                 :fill-pointer 0)))
       (maphash #'(lambda (key val)
                    (vector-push-extend (list key val)
                                        dump-list))
                hash-table)
       (coerce dump-list 'list)))

